const mfb = @import("minifb");
const std = @import("std");
const min = std.math.min;
const max = std.math.max;

const Width = 800;
const Height = 600;

fn grey(value: u8) u32 {
    return mfb.rgb(value, value, value);
}

const Coords = struct {
    x: i32,
    y: i32,
};

const State = struct {
    period: f64,
    bgValue: u8 = 0,
    increasing: bool = true,
    timer: mfb.Timer,

    pub fn init(period: f64) !State {
        const timer = try mfb.Timer.init();
        return State{
            .period=period,
            .timer=timer,
        };
    }

    pub fn deinit(self: State) void {
        self.timer.deinit();
    }

    pub fn render(self: State, buf: Buffer) void {
        var bgValue = @rem(self.timer.now(), self.period) / self.period * 512;
        if (bgValue >= 256) {
            bgValue = 512 - bgValue;
        }

        buf.drawRectangle(.{.x=0,.y=0}, .{.x=Width, .y=Height}, grey(@floatToInt(u8, bgValue)));
        buf.drawRectangle(.{.x=Width/3, .y=Height/3}, .{.x=2*Width/3, .y=2*Height/3}, mfb.rgb(255,0,0));
    }
};

const Buffer = struct {
    alloc: *std.mem.Allocator,
    slice: []u32,
    width: u32,
    height: u32,

    pub fn init(alloc: *std.mem.Allocator, width: u32, height: u32) !Buffer {
        const slice = try alloc.alloc(u32, width*height);
        return Buffer {
            .alloc=alloc,
            .slice=slice,
            .width=width,
            .height=height,
        };
    }

    pub fn deinit(self: Buffer) void {
        self.alloc.free(self.slice);
    }

    pub fn setPixel(self: Buffer, coords: Coords, color: u32) void {
        const x = @intCast(u32, coords.x);
        const y = @intCast(u32, coords.y);
        self.slice[y*self.width + x] = color;
    }

    pub fn drawRectangle(self: Buffer, a: Coords, b: Coords, color: u32) void {
        var start = Coords{
            .x=max(0, min(a.x, b.x)),
            .y=max(0, min(a.y, b.y)),
        };
        var end = Coords{
            .x=min(@intCast(i32, self.width), max(a.x, b.x)),
            .y=min(@intCast(i32, self.height), max(a.y, b.y)),
        };

        var y = start.y;
        while (y < end.y) {
            var x = start.x;
            while (x < end.x) {
                self.setPixel(.{.x=x, .y=y}, color);
                x += 1;
            }
            y += 1;
        }
    }
};

pub fn main() !void {
    var gp_allocator = std.heap.GeneralPurposeAllocator(.{}){};
    const alloc = &gp_allocator.allocator;

    std.log.info("All your codebase are belong to us.", .{});
    var win = mfb.Window.open("Hello minifb-zig", Width, Height) catch unreachable;
    mfb.setTargetFPS(7);

    var state = try State.init(3);
    defer state.deinit();

    var buf = try Buffer.init(alloc, Width, Height);
    defer buf.deinit();

    state.render(buf);
    while (win.waitSync()) {
        state.render(buf);
        // TODO Handle mfb state
        _ = win.update(buf.slice) catch unreachable;
    }
}
